#!/usr/bin/env python2.6

'''Little UI to scan from an HP CM1312nfi MFP.

Usage: $0 directory basename

Will put all the images into "directory" and prefix them with "basename".

You can start the scanner by pressing the "Start scan" on the scanner
of by clicking on the "Scan" button in the UI.
'''

import sys
import os
import re

sys.path.insert(0, '/usr/share/hplip')
#print sys.path

import scanext
import sane
from threading import Thread, Event
from Queue import Queue, Empty
import Image
import StringIO
from PyQt4.QtGui import *
from PyQt4.QtCore import *
from gallery import GalleryMain, PixmapView, loaderThread, KDEThumbnailCache
import httplib
import traceback
from types import UnicodeType
from ConfigParser import SafeConfigParser

from base.g import log
if 0:
    log.set_level('debug')

VERSION="0.6 (23.07.2009)"

#print sane.get_devices()

# TODO
# - Dialog to edit project settings
# - PyScan hangs when scanner is turned off before exiting
# - Scan hangs when there is an error during scanning (abort)
# - Allow to scan a page again
# - Delete one or more pages
# - Insert a page
# - Renumber a page range
# - Reorder pages by drag'n'drop
# - Select a recent project
# - Support for double-sided batch scanning
# - Ctrl-Q to quit
# - Call hugin to join double-page images
# - Show ADF status in UI

class SaveThread(Thread):
    saved = SIGNAL('saved') # filename
    
    def __init__(self):
        super(SaveThread,self).__init__()
        
        self.queue = Queue()
        self._run = True
        self.qt = QObject()
    
    def run(self):
        while self._run:
            job = self.queue.get(True)
            if job is None:
                break
                
            img, project = job
            n, name = project.next()
            name = os.path.abspath(name)
            path, basename = os.path.split(name)
            tmp = os.path.join(path, '.%s_' % basename)
            print repr('Saving image %s...' % tmp)
            img.save(tmp, project.format, project.quality)
            os.rename(tmp, name)
            
            self.qt.emit(self.saved, name, img)
    
    def stop(self):
        if self.isAlive():
            self._run = False
            self.queue.put(None)
            self.join(60)
    
    def save(self, im, project):
        self.queue.put((im, project))

class ImageBuffer(object):
    def __init__(self, width, height, bytesPerLine, format):
        self.width, self.height, self.bytesPerLine = width, height, bytesPerLine
        self.format = format
        self.size = self.height * self.bytesPerLine
        print self.width, self.height, self.bytesPerLine, self.size, self.format
        
        self.buffer = QByteArray(self.size, '\x00')
        self.offset = 0
        print 'ImageBuffer size:', self.buffer.size()
    
    def write(self, data):
        l = len(data)
        self.buffer.replace(self.offset, l, data)
        self.offset += l
    
    def seek(self, offset):
        #self.offset = offset
        pass
    
    def toImage(self):
        # Data is written as RGBA but on an little endian machine, that
        # becomes ARGB when you access it as int.
        #self.data = str(self.buffer)
        b = str(self.buffer)
        #print len(b),self.width, self.height,self.bytesPerLine
        img = QImage(b, self.width, self.height, 
            self.bytesPerLine, self.format)
        img.b = b
        #print img, img.isNull()
        return img

class ScannerThread(Thread):
    scanStarted = SIGNAL('scanStarted') # self
    newImageAvailable = SIGNAL('newImageAvailable') # self, im
    progressSignal = SIGNAL('progress') # self, im
    
    def __init__(self, dev):
        super(ScannerThread,self).__init__()
        
        self.dev = dev
        self.autoSource = False
        self.moreDocuments = False
    
        if 0:
            print dev.getParameters()
        
            for opt in dev.getOptions():
                print opt

            print dev.getOption('optlist')
        
        self.qt = QObject()
        
        self.im = None
        self.queue = Queue()
        self.updateQueue = Queue()
        self.eventQueue = Queue()
        self.conn = httplib.HTTPConnection(dev.ip)
        
        self.startScanPattern = re.compile(r'<StartScan>(\d)+')
        self.adfLoadedPattern = re.compile(r'<ADFLoaded>(\d)+')
	self.adfLoaded = False
	self.scanButtonPressed = False
        
    def emit(self, signal, *args):
        self.qt.emit(signal, *args)
    
    def refreshScanOptions(self):
        url = '/hp/device/notifications.xml'
        self.conn.request('GET', url)
        resp = self.conn.getresponse()
        if resp.status != httplib.OK:
            raise Exception('Scanner returned %d %s' % (resp.status, resp.reason))
        
        data = resp.read()
        #print data
        
        m = self.adfLoadedPattern.search(data)
        #print 'adf',m.group(1)
	oldAdfLoaded = self.adfLoaded
        self.adfLoaded = int(m.group(1))
	if oldAdfLoaded != self.adfLoaded:
	    print 'adf',self.adfLoaded

        if self.autoSource:
            self.dev.setOption('source', 'ADF' if self.adfLoaded else 'Flatbed')
        m = self.startScanPattern.search(data)
        #print 'scan',m.group(1)
	old = self.scanButtonPressed
        self.scanButtonPressed = int(m.group(1))
	if old != self.scanButtonPressed:
	    print 'scanButtonPressed',self.scanButtonPressed
    
    def scan(self):
        self.queue.put('Scan')
    
    def createImageBuffer(self):
        "Snap a picture, returning a PIL image object with the results"
        (format, mode, last_frame,
         width, height, depth, bytes_per_line) = self.dev.getParameters()
        if mode in ['gray', 'red', 'green', 'blue']:
            raise Exception('Unsupported format %s' % mode)
        elif mode == 'color':
            format = QImage.Format_RGB888
        else:
            raise ValueError('got unknown "mode" from self.getParameters()')
        self.imgBuffer = ImageBuffer(width, height, bytes_per_line, format)
        return self.imgBuffer
    
    def run(self):
        while True:
            try:
                self.refreshScanOptions()
                if self.scanButtonPressed or self.moreDocuments:
                    self.scan()
            except:
                traceback.print_exc()
            
            # Wait for start signal
            try:
                s = self.queue.get(True, 1)
            except Empty:
                continue
            
            if not s:
                break
            
            try:
                self.scanOneFrame()
            except:
                traceback.print_exc()
    
    def stop(self):
        if self.isAlive():
            self.queue.put(None)
            self.join(1)
    
    def scanOneFrame(self):
        print 'Getting scan'
        needsCancel = False
        self.moreDocuments = False
        try:
            # TODO Delete
            self.createImageBuffer()
            self.im = QImage()
            print 'emit scanStarted'
            self.emit(self.scanStarted, self, self.im)
            
            try:
                needsCancel = True
                result = self.dev.startScan(self.imgBuffer, self.updateQueue, self.eventQueue)
                print result
                # TODO Delete
                #self.dev.dev.snap(self.im.im.id, 1)
		#self.dev.waitForScanActive()
                self.progress()
            except scanext.error, v:
                print v
                print repr(v)
                if v == 'Document feeder out of documents':
                    pass
                elif v == 'Operation was cancelled':
                    pass
                else:
                    raise Exception(v)
            
            #print 'Done.'
        finally:
            self.refreshScanOptions()
            if self.adfLoaded:
                self.moreDocuments = True
                return
            
            self.dev.dev.cancelScan()
    
    def progress(self):
        print 'progress'
        last = 0
        self.dev.waitForScanActive()
        while self.dev.isScanActive():
            try:
                status, bytes_read = self.updateQueue.get(True, 5)
            except Empty:
                break
            
            while self.updateQueue.qsize():
                status, bytes_read = self.updateQueue.get()
            
            p = int(self.imgBuffer.offset * 100 / self.imgBuffer.size)
            if p != last:
                last = p
                sys.stdout.write('progress %s %9d %2d\r' % (repr(status), bytes_read, p))
                sys.stdout.flush()
                self.im = self.imgBuffer.toImage()
                #print 'progress',self.im,self.im.isNull()
                self.emit(self.progressSignal, self, self.im, p)
                
            if self.imgBuffer.offset == self.imgBuffer.size:
                self.im = self.imgBuffer.toImage()
                self.emit(self.progressSignal, self, self.im, 100)
                print 'emit newImageAvailable'
                self.emit(self.newImageAvailable, self, self.im)
                break
        
        self.dev.waitForScanDone()
        print 'Scan finished. %d' % self.imgBuffer.offset

def pilImage2QImage(im, width=None, height=None):
    if im is None:
        return QPixmap()
    
    if 0: # Uses less memory but slow or ugly
        if width or height:
            if width and not height:
                height = width
            elif height and not width:
                width = height
            
            im = im.copy()
            # Slow :(
            #im.thumbnail((width, height), Image.ANTIALIAS)
            # Ugly :(
            #im.thumbnail((width, height), Image.BILINEAR)
            im.thumbnail((width, height), Image.BICUBIC)
    
    b = im.convert('RGB').tostring()
    #print len(b), repr(b[:32])
    w, h = im.size
    #print w*h*3
    #print w*h*4
    bytesPerLine = len(b)//h
    img = QImage(b, w, h, bytesPerLine, QImage.Format_RGB888)
    img.b = b
    
    if 0: # Needs more RAM but fast
        if width and height:
            img = img.scaled(width, height,
                Qt.KeepAspectRatio,
                Qt.SmoothTransformation)
        elif height:
            img = img.scaledToHeight(height, Qt.SmoothTransformation)
        elif width:
            img = img.scaledToWidth(width, Qt.SmoothTransformation)
    
    return scaleQImage(img, width, height)

def scaleQImage(img, width=None, height=None):
    # Needs more RAM but fast
    if width and height:
        img = img.scaled(width, height,
            Qt.KeepAspectRatio,
            Qt.SmoothTransformation)
    elif height:
        img = img.scaledToHeight(height, Qt.SmoothTransformation)
    elif width:
        img = img.scaledToWidth(width, Qt.SmoothTransformation)
    
    return img
    
def join(a, *p):
    """Join two or more pathname components, inserting '/' as needed.
    If any component is an absolute path, all previous path components
    will be discarded."""
    path = a
    for b in p:
        if b.startswith(u'/'):
            path = b
        elif path == u'' or path.endswith(u'/'):
            path +=  b
        else:
            path += u'/' + b
    return path

os.path.join = join

class Main(GalleryMain):
    def __init__(self, dev, project, parent=None):
        super(Main,self).__init__(parent)
        self.resize(1200,600)
        
        self.dev = dev
        self.project = project
        self.project.updateSaneDevice(self.dev)
        
        self.toolbar = QToolBar()
        self.addToolBar(self.toolbar)
        
        self.currentScan = PixmapView('Scan', self)
        self.splitter.addWidget(self.currentScan)
        self.splitter.setStretchFactor(2, 1)
        
        self.timer = QTimer()
        self.timer.setInterval(1000)
        self.connect(self.timer, SIGNAL('timeout()'), self.tick)
        
        self.setFiles(self.project.listdir())
        
        self.scanner = False
        
        self.saveThread = SaveThread()
        self.connect(self.saveThread.qt, SaveThread.saved, self.saved)
        
        self.scannerThread = ScannerThread(self.dev)
        self.scannerThread.autoSource = self.project.autoSource
        self.connect(self.scannerThread.qt, ScannerThread.newImageAvailable, self.newImageAvailable)
        self.connect(self.scannerThread.qt, ScannerThread.scanStarted, self.scanStarted)
        self.connect(self.scannerThread.qt, ScannerThread.progressSignal, self.refreshCurrentScan)
        
        self.createActions()
        self.toolbar.addAction(self.startScanAction)
    
    def startThreads(self):
        self.saveThread.start()
        self.scannerThread.start()
    
    def stopThreads(self):
        self.saveThread.stop()
        self.scannerThread.stop()
    
    def createActions(self):
        self.startScanAction = QAction(u'&Scan', self)
        self.connect(self.startScanAction, SIGNAL('triggered(bool)'), self.startScan)
    
    def startScan(self, checked=False):
        self.scannerThread.scan()
    
    def scanStarted(self, st, im):
        print 'scanStarted'
	self.scanImage = im
        self.scanner = True
        self.filename = None
        self.startScanAction.setEnabled(False)
    
    def newImageAvailable(self, st, im):
        print 'newImageAvailable'
        self.scanner = False
	self.scanImage = None
        self.refreshCurrentScan(None, im, 100)
        self.startScanAction.setEnabled(True)
        
        self.saveThread.save(im, self.project)
    
    def saved(self, filename, im):
        print 'saved',repr(filename)
        # TODO Delete
        #img = pilImage2QImage(im)
        loaderThread.register(filename, im)
        
        model = self.thumbnails.model
        n = model.add(filename)
        index = model.index(n, 0)
        if index.isValid():
            self.thumbnails.selectionModel().setCurrentIndex(index,
                QItemSelectionModel.ClearAndSelect)
    
    def refreshCurrentScan(self, scan, im, p):
	if not self.scanner:
	    return

        #print 'refreshCurrentScan',p
        #im = self.scannerThread.im
	#im = self.scanImage
        size = self.currentScan.scrollArea.viewport().size()
        # TODO Delete
        #img = pilImage2QImage(im, size.width(), size.height())
        im = scaleQImage(im, size.width(), size.height())
        pm = QPixmap.fromImage(im)
        self.currentScan.setPixmap(pm)
    
    def show(self):
        self.timer.start()
        QMainWindow.show(self)
    
    def tick(self, *args):
        s = 'Inactive'
        if self.scanner:
            #self.refreshCurrentScan()
            s = u'Scanning...'
        elif self.currentImage.filename is not None:
            s =  unicode(self.currentImage.filename)
        
        self.setWindowTitle('PyScan %s - %s' % (VERSION, s))

class ScanProject(object):
    def __init__(self, path, prefix=u'Seite', format='TIFF'):
        self.path = os.path.abspath(path)
        self.mode = 'Color'
        self.resolution = 300
        self.source = 'Auto'
        self.format = format
        self.scanArea = (0.0, 0.0, 215.9, 296.92)
        self.quality = 90
        self.start = 1
        self.step = 1
        self.prefix = prefix
        
        self.postInit()
        
    def postInit(self):
        if self.prefix[-1] not in '.-_':
            self.prefix += '-'
        
        if self.format == 'JPEG':
            self.suffix = u'.jpg'
        else:
            self.suffix = u'.%s' % (self.format.lower())
        
        self.autoSource = self.source == 'Auto'
        self.refresh()
    
    def createConfig(self):
        return SafeConfigParser({
            'mode': 'Color',
            'resolution': '300',
            'source': 'Auto',
            'format': 'TIFF',
            'scanArea': '0.00, 0.00, 215.90, 296.92',
            'quality': "90",
            'start': '1',
            'step': '1',
        })
    
    def save(self):
        filename = os.path.join(self.path, ".pyScan")
        
        config = self.createConfig()
        if os.path.exists(filename):
            config.read(filename)
        
        section = 'PyScan'
        if not config.has_section(section):
            config.add_section(section)
        
        config.set(section, 'prefix', self.prefix)
        config.set(section, 'step', unicode(self.step))
        config.set(section, 'start', unicode(self.start))
        config.set(section, 'quality', unicode(self.quality))
        area = u', '.join(map(lambda f: u'%.2f' % f, self.scanArea))
        config.set(section, 'scanArea', area)
        config.set(section, 'format', self.format)
        config.set(section, 'source', self.source)
        config.set(section, 'resolution', unicode(self.resolution))
        config.set(section, 'mode', self.mode)
        
        with open(filename, 'wb') as configfile:
            config.write(configfile)
    
    def load(self):
        filename = os.path.join(self.path, ".pyScan")
        if not os.path.exists(filename):
            return
        
        config = self.createConfig()
        config.read(filename)
        
        section = 'PyScan'
        self.mode = config.get(section, 'mode')
        self.resolution = config.getint(section, 'resolution')
        self.source = config.get(section, 'source')
        self.format = config.get(section, 'format')
        area = config.get(section, 'scanArea')
        values = []
        for v in area.split(r','):
            v = float(v)
            values.append(v)
        self.scanArea = tuple(values)
        self.quality = config.getint(section, 'quality')
        self.start = config.getint(section, 'start')
        self.step = config.getint(section, 'step')
        self.prefix = config.get(section, 'prefix')
        
        self.postInit()
    
    def dump(self):
        for key, value in self.__dict__.items():
            if key in ('files',):
                continue
            
            if key == 'scanArea':
                print key, '%.2f, %.2f, %.2f, %.2f' % self.scanArea 
            else:
                print key, value
    
    def isAutoSource(self):
        return self.source == 'Auto'
    
    def updateSaneDevice(self, dev):
        dev.setOption('mode', self.mode)
        dev.setOption('resolution', self.resolution)
        if not self.autoSource:
            dev.setOption('source', self.source)
        dev.setOption('tl_x', self.scanArea[0])
        dev.setOption('tl_y', self.scanArea[1])
        dev.setOption('br_x', self.scanArea[2])
        dev.setOption('br_y', self.scanArea[3])
    
    def add(self, filename):
        self.files.append(filename)
        self.files.sort()
    
    def refresh(self):
        self.files = None
        self.listdir()
    
    def listdir(self):
        if self.files:
            return self.files
        
        pattern = re.compile(re.escape(self.prefix)+'(\d{4})\.[a-zA-Z]{3,4}')
        
        l = []
        for f in os.listdir(self.path):
            m = pattern.match(f)
            if m:
                l.append(f)
        
        l.sort()
        self.files = [os.path.join(self.path, f) for f in l]
        
        return self.files
    
    def getFilename(self, n):
        return os.path.join(self.path, u'%s%04d%s' % (self.prefix, n, self.suffix))
    
    def last(self):
        l = self.listdir()
        return l[-1] if l else None
    
    def next(self):
        n = self.start
        
        while True:
            filename = self.getFilename(n)
            if not os.path.exists(filename):
                break
            n += self.step
        
        return n, filename

    def insert(self, pos, count=1):
        last, filename = self.next()
        kdeCache = KDEThumbnailCache(1)
        
        x = last
        
        # Can't use >/< here since step can be positive or negative
        pos -= self.step
        offset = count * self.step
        while x != pos:
            oldName = self.getFilename(x)
            newName = self.getFilename(x+offset)
            
            if os.path.exists(newName):
                raise 'Insert failed at %d: File %s already exists' % (x, newName)
            
            # Mind gaps
            if os.path.exists(oldName):
                print 'mv',oldName,newName
                os.rename(oldName, newName)
            
            oldCachePath = kdeCache.getCachePath(oldName)
            newCachePath = kdeCache.getCachePath(newName)
            if os.path.exists(newCachePath):
                os.remove(newCachePath)
            if os.path.exists(oldCachePath):
                os.rename(oldCachePath, newCachePath)
                os.utime(newCachePath, None)
            
            x -= self.step
    
    def delete(self, pos, count=1):
        last, filename = self.next()
        kdeCache = KDEThumbnailCache(1)
        
        x = pos
        offset = count * self.step
        while x != last:
            destName = self.getFilename(x)
            sourceName = self.getFilename(x + offset)
            
            if os.path.exists(destName):
                os.remove(destName)
            if os.path.exists(sourceName):
                os.rename(sourceName, destName)
            
            x += self.step
    
    def move(self, start, count, dest):
        last, filename = self.next()
        # TODO First, create a gap for the files to move
        # TODO Move the files to their new home
        # TODO Delete the gap

if __name__ == '__main__':
    if 1:
        app = QApplication(sys.argv)

	sane.init()
        
        ip = '192.168.1.242'
        uri = 'hpaio:/net/HP_Color_LaserJet_CM1312nfi_MFP?ip=%s' % ip
        try:
            dev = sane.ScanDevice(uri)
        except TypeError, v:
            if str(v).startswith('__init__() takes at most'):
                raise Exception("It seems you didn't apply the patch in sane-py.patch (see README)")
            else:
                raise
        except scanext.error, v:
            if str(v) == 'Error during device I/O':
                raise Exception("Can't access scanner. Is it switched on?")
            raise
        dev.ip = ip
        dev.compression = 'None'

        print dir(dev.dev)

        #filePattern = sys.argv[1]
        #start = int(sys.argv[2])
        #inc = int(sys.argv[3])
        
        path = unicode(sys.argv[1], 'utf-8')
        if not os.path.exists(path):
            os.makedirs(path)
        elif not os.path.isdir(path):
            raise Exception('%s ist kein Verzeichnis' % path)
        
        prefix = 'Seite'
        if len(sys.argv) >= 3:
            prefix = unicode(sys.argv[2], 'utf-8')
        format = 'TIFF'
        if len(sys.argv) >= 4:
            format = sys.argv[3].upper()
            if format == 'JPG':
                format = 'JPEG'
        
        #print repr(path),repr(prefix),repr(format)
        project = ScanProject(path, prefix=prefix, format=format)
        project.start = 1
        project.step = 1
        
        project.load()
        project.dump()
        
        if 1:
            main = Main(dev, project)
            main.startThreads()
            #main.setScanner(dev, filePattern, start, inc)
            main.show()
            
            rc = app.exec_()
            main.stopThreads()
            dev.closeScan()
            
            project.save()
            
            sys.exit(rc)
        
        if 0:
            project.insert(374, 1)
    
    if 0:
        im = Image.open('../scanner.bak/test.tiff')
        img = pilImage2QImage(im)
        img.save('qt-test.tiff', 'TIFF')
